/***************************************************************************
 *   Copyright (C) 2004 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#include "arc.h"

#include <cmath>

namespace Papyrus {

Arc::Arc( double radius,
          double angle_start,
          double angle_end,
          Cairo::RefPtr<Cairo::Pattern> fill
        ) :
    Shape(fill),
    m_radius(radius),
    m_angle_start(angle_start),
    m_angle_end(angle_end) {
        if (angle_start > angle_end) {
            m_angle_start = angle_start;
            m_angle_end = angle_end;
        }

        this->calculate_extents();
    }

    Arc::Arc( double radius,
              double angle_start,
              double angle_end,
              Cairo::RefPtr<Cairo::Pattern> fill,
              const LineStyle& outline
            ) :
    Shape(fill, outline),
    m_radius(radius),
    m_angle_start(angle_start),
    m_angle_end(angle_end) {
      if (angle_start > angle_end) {
        m_angle_start = angle_start;
        m_angle_end = angle_end;
      }

      this->calculate_extents();
    }

    Arc::pointer Arc::create( double radius,
                              double angle_start,
                              double angle_end,
                              Cairo::RefPtr<Cairo::Pattern> fill
                            )
    {
        return Arc::pointer( new Arc(radius, angle_start, angle_end, fill) );
    }

    Arc::pointer Arc::create( double radius,
                              double angle_start,
                              double angle_end,
                              Cairo::RefPtr<Cairo::Pattern> fill,
                              const LineStyle& outline
                            )
    {
        return Arc::pointer( new Arc(radius, angle_start, angle_end, fill, outline) );
    }

    Arc::~Arc() {}

    double Arc::radius() {
        return m_radius;
    }

    void Arc::set_radius(double r) {
        m_radius = r;
        this->shape_changed();
        m_signal_radius.emit();
    }

    void Arc::draw_shape( Cairo::RefPtr<Cairo::Context> cairo ) {
        cairo->arc( 0.0, 0.0, m_radius, m_angle_start, m_angle_end);
        cairo->line_to( 0.0, 0.0);
    }

    void Arc::draw_outline( Cairo::RefPtr< Cairo::Context > cairo ) {
        cairo->arc( 0.0, 0.0, m_radius, m_angle_start, m_angle_end);
        if (m_angle_end - m_angle_start < 2.0*M_PI) {
            cairo->line_to( 0.0, 0.0);
            cairo->line_to( cos(m_angle_start)*m_radius, sin(m_angle_start)*m_radius);
        }
    }

    double Arc::start_angle( ) {
        return m_angle_start;
    }

    double Arc::end_angle( ) {
        return m_angle_end;
    }

    void Arc::set_angles( double start, double end ) {
        if (start < end) {
            m_angle_start = start;
            m_angle_end = end;
        } else {
            m_angle_start = end;
            m_angle_end = start;
        }
        this->shape_changed();
        m_signal_angle.emit();
    }

    void Arc::need_redraw() {
        this->calculate_extents();
        Drawable::need_redraw();
    }

    void Arc::calculate_extents( ) {
        Region temp_region;

        temp_region.set_xywh(0.0, 0.0, 0.0, 0.0);

        temp_region.include(cos(m_angle_start)*m_radius, sin(m_angle_start)*m_radius);
        temp_region.include(cos(m_angle_end)*m_radius, sin(m_angle_end)*m_radius);
        if (m_angle_start <= M_PI_2 && m_angle_end >= M_PI_2)
            temp_region.include(0.0, m_radius);
        if (m_angle_start <= M_PI && m_angle_end >= M_PI)
            temp_region.include(-m_radius, 0.0);
        if (m_angle_start <= 3.0*M_PI_2 && m_angle_end >= 3.0*M_PI_2)
            temp_region.include(0.0, -m_radius);
        if (m_angle_start <= 2.0*M_PI && m_angle_end >= 2.0*M_PI)
            temp_region.include(m_radius, 0.0);

        this->set_extents( temp_region );
    }

    sigc::signal<void> Arc::signal_radius() {
        return m_signal_radius;
    }

    sigc::signal<void> Arc::signal_angle() {
        return m_signal_angle;
    }

}
