/***************************************************************************
 *   Copyright (C) 2004 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#ifndef PAPYRUSDRAWABLE_H
#define PAPYRUSDRAWABLE_H

#include <papyrus/smart_pointer.h>
#include <papyrus/utility.h>

#include <papyrus/renderable.h>
#include <papyrus/region.h>
#include <papyrus/viewbox.h>

namespace Papyrus
  {

  class Group;

  /**
   * Papyrus::Drawable is the base class for objects capable of being drawn into a cairo context that
   * also have spatial properties; ie affine transforms can be meaningfully applied to
   * drawables. It is the presence of meaningful affine transforms that differentiates
   * Papyrus::Drawable from its parent Papyrus::Renderable.
   *
   * However Papyrus::Drawable lacks a concept of filling and outlining which differentiates
   * it from its child Papyrus::Shape.
   *
   * This class provides a common interface for managing a transformation matrix that
   * is applied before the object is drawn. These transformations include translation,
   * scaling and shearing (skewing) in the x and y dimensions as well as rotation in
   * the x-y plane.
   *
   * This class implements the render method required of the parent
   * renderable by pushing the current cairo state onto the cairo
   * stack, applying any necessary transformation matrix, calling
   * the pure virtual draw method, and popping the cairo stack to
   * restore the state.
   *
   * Since this class handles the matrices directly, children only
   * need to override the draw method and draw themselves with a
   * local coordinate frame reference.
   *
   * When inheriting from Papyrus::Drawable you should provide a child specific implementation
   * of the draw() method. You should probably also override the extents() method to return
   * a Region object that is meaningful to your child, along with providing a child specific
   * implementation of intersects(x,y) to provide a meaningful (and efficient) method of
   * determining whether your child includes the point (x,y).
   *
   * @author Rick L Vinyard Jr
   */
  class Drawable : public Renderable
    {
    public:
      typedef PapyrusSmartPointer<Drawable> pointer;

      /**
       * Constructor that sets initial transforms.
       *
       * If no parameters are specified, initial values set translation to (0,0) with a (1,1) scale
       * ratio and no rotation or shearing.
       *
       * If either scale parameter < 0.0, the parameter will be reset to 1.0.
       */
      Drawable( double tx = 0.0, double ty = 0.0, double sx = 1.0, double sy = 1.0, double r = 0.0 );

      virtual ~Drawable();

      bool is_visible();

      void show();

      void hide();

      /** X position where this Drawable will be drawn */
      double get_translate_x();

      /** Y position where this Drawable will be drawn */
      double get_translate_y();

      /** (X,Y) position where this Drawable will be drawn */
      void get_translate( double& tx, double& ty );

      /**
       * Sets the X position
       *
       * If previous position was (px,py) new position is (tx,py).
       */
      void set_translate_x( double tx );

      /**
       * Sets the Y position
       *
       * If previous position was (px,py) new position is (px,ty).
       */
      void set_translate_y( double ty );

      /**
       * Sets position to (X,Y)
       *
       * Absolute move to (tx,ty).
       */
      void set_translate( double tx, double ty );

      /**
       * Translates current position by tx
       *
       * If previous position was (px,py) new position is (px+tx,py).
       */
      void translate_x( double tx );

      /**
       * Translates current position by ty
       *
       * If previous position was (px,py) new position is (px,py+ty).
       */
      void translate_y( double ty );

      /**
       * Translate current position by (tx, ty)
       *
       * If previous position was (px,py) new position is (px+tx, py+ty).
       */
      void translate( double tx, double ty );

      /** returns the current value of the x scaling */
      double get_scale_x();

      /** returns the current value of the y scaling */
      double get_scale_y();

      /** convenience method to get scaling in x and y dimensions with a single function call */
      void get_scale( double& sx, double& sy );

      /**
       * sets scaling in the x dimension and emits changed signal with parameter SCALE_X
       *
       * @param scale_x The new x dimension scaling; must be >= 0.0 or no change will occur
       */
      void set_scale_x( double sx );

      /**
       * sets scaling in the y dimension and emits changed signal with parameter SCALE_Y
       *
       * @param scale_y The new y dimension scaling; must be >= 0.0 or no change will occur
       */
      void set_scale_y( double sy );

      /**
       * sets scaling in the x and y dimensions and emits changed signal with one of parameters SCALE_X, SCALE_Y or SCALE_X|SCALE_Y
       *
       * @param scale_x The new x dimension scaling; must be >= 0.0 or no change to x scaling will occur
       * @param scale_y The new y dimension scaling; must be >= 0.0 or no change to y scaling will occur
       */
      void set_scale( double scale_x, double scale_y );

      /**
       * sets scaling in the x and y dimensions and emits changed signal with parameter SCALE_X|SCALE_Y
       *
       * @param s The amount to scale the x and y dimensions by; must be >= 0.0 or no change will occur
       */
      void set_scale( double s );

      /**
       * Scales the current x dimension by s and emits changed signal with parameter SCALE_X
       *
       * @param s The amount to scale the x dimension by; must be >= 0.0 or no change will occur
       */
      void scale_x( double s );

      /**
       * Scales the current y dimension by s and emits changed signal with parameter SCALE_Y
       *
       * @param s The amount to scale the y dimension by; must be >= 0.0 or no change will occur
       */
      void scale_y( double s );

      /**
       * Scales the current x and y dimensions by s and emits changed signal with parameter SCALE_X|SCALE_Y
       *
       * @param s The amount to scale the x and y dimensions by; must be >= 0.0 or no change will occur
       */
      void scale( double s );

      /**
       * Scales the current x dimension by sx and the y dimension by sy and emits changed signal with one of parameters SCALE_X, SCALE_Y or SCALE_X|SCALE_Y
       *
       * @param scale_x The amount to scale the x dimension by; must be >= 0.0 or no change will occur
       * @param scale_y The amount to scale the y dimension by; must be >= 0.0 or no change will occur
       */
      void scale( double scale_x, double scale_y );

      /** Get rotation r in radians */
      double get_rotate();

      /** Rotate to r radians */
      void set_rotate( double r );

      /**
       * Rotate by r radians
       *
       * If previous rotation was pr then new rotation is pr+r
       */
      void rotate( double r );

      /**
       * Get x skewing angle in radians
       *
       * The skew angle is used to implement a shearing transform with the shear value = tan(skewx).
       */
      double get_skew_x();

      /**
       * Set x skewing angle in radians
       *
       * The skew angle is used to implement a shearing transform with the shear value = tan(skewx).
       */
      void set_skew_x(double skewx);

      /**
       * Skew the Drawable in the x direction by skewx radians
       *
       * If previous x skew angle was px, new skew angle is px+skewx.
       */
      void skew_x(double skewx);

      /**
       * Get y skewing angle in radians
       *
       * The skew angle is used to implement a shearing transform with the shear value = tan(skewy).
       */
      double get_skew_y();

      /**
       * Set y skewing angle in radians
       *
       * The skew angle is used to implement a shearing transform with the shear value = tan(skewy).
       */
      void set_skew_y(double skewy);

      /**
       * Skew the Drawable in the y direction by skewy radians
       *
       * If previous y skew angle was py, new skew angle is py+skewy.
       */
      void skew_y(double skewy);

      /** Get the x and y skewing angles */
      void get_skew(double& skewx, double& skewy);

      /** Set the x and y skewing angles */
      void set_skew(double skewx, double skewy);

      /** Skew the x and y angles uniformly */
      void skew(double s);

      /** Skew the x and y angles independently */
      void skew(double skewx, double skewy);

      /** Get centroid x position about which the Drawable is rotated */
      double get_centroid_x();

      /** Set the x value of the centroid about which the Drawable will be rotated */
      void set_centroid_x(double centroid_x);

      /** Get centroid y position about which the Drawable is rotated */
      double get_centroid_y();

      /** Set the y value of the centroid about which the Drawable will be rotated */
      void set_centroid_y(double centroid_y);

      /** Get the (x,y) centroid about which the Drawable will be rotated */
      void get_centroid(double& centroid_x, double& centroid_y);

      /** Set the (x,y) centroid about which the Drawable will be rotated */
      void set_centroid(double centroid_x, double centroid_y);

      /**
       * Reset position to default values of xy = (0,0), scale = (1,1),
       * rotation = 0, and shear xy = (0,0).
       */
      void reset_position();

      /** Return the cairo matrix of this object with all transforms applied */
      const Cairo::Matrix& matrix();

      /** Returns the global matrix that will be applied to this drawable. */
      Cairo::Matrix global_matrix();

      /** Transform a global point to the drawable's local coordinates */
      void global_to_local( double& x, double& y );

      /** Transform a local point to global coordinates */
      void local_to_global( double& x, double& y );

      /** Transform a global distance to the drawable's local coordinates */
      void distance_global_to_local( double& x, double& y );

      /** Transform a local distance to global coordinates */
      void distance_local_to_global( double& x, double& y );

      /**
       * Reimplemented render method from renderable.
       *
       * Generally, children should reimplement the draw method rather than this
       * method.
       *
       * Performs the following actions:
       *   1. Pushes cairo state onto the stack
       *   2. Recalculates local matrix if necessary
       *   3. Applies the local matrix to the current matrix
       *   4. Calls the virtual draw method
       *   5. Restores the cairo state
       */
      virtual void render( Cairo::RefPtr<Cairo::Context> cairo );

      /**
       * Pure virtual draw method to be reimplemented by children.
       *
       * This method is called by the supplied render method after the local
       * transformation matrix has been applied to the cairo stack.
       */
      virtual void draw( Cairo::RefPtr<Cairo::Context> cairo ) = 0;

      /**
       * Returns the extents of this drawable \b without applying the matrix formed from the
       * affine transform methods translate*, scale* and rotate*.
       *
       * Children should modify the member attribute m_extents as appropriate and
       * generally shouldn't need to provide a polymorphic version of this method.
       */
      virtual Region& extents();

      virtual Region local_extents();

      /**
       * Returns the extents of this drawable that has been transformed by the matrix formed
       * by the affine transform methods translate*, scale* and rotate*.
       *
       * Applies the local transformation matrix to the extents returned from the
       * extents() method. The m_extents member is not used so that the
       * extents() method may be overridden by a child and this method will still work as
       * intended without also overriding this method.
       */
      virtual Region global_extents();

      /**
       * Returns the extents of this drawable that has been transformed by a combined local
       * and global matrix.
       *
       * The local matrix is applied to the supplied parameter matrix, and this transform is in
       * turn applied to the extents returned from extents().
       *
       * The m_extents member is not used so that the extents() method may be
       * overridden by a child and this method will still work as intended without also overriding
       * this method.
       */
      virtual Region extents_transformed( Cairo::Matrix& globalmatrix );

      /**
       * True if the point (x,y) is inside this Drawable's drawing area.
       */
      virtual bool inside( double x, double y );

      bool exclude_from_extents();

      void set_exclude_from_extents(bool b);

      /**
       * Set to true if this Drawable will respond to the select method.
       *
       * For example, it may not be desirable to have a group respond to the
       * select method itself, and may instead pass the select(x,y) request
       * to its children.
       */
      void set_selectable( bool selectable = true );

      /**
       * True if this Drawable will respond to the select method, or false if
       * it will not include itself in the select method.
       */
      bool is_selectable();

      /** True if this Drawable contains other drawables. */
      virtual bool is_group();

      /**
       * Returns the viewbox associated with this Drawable.
       *
       * If no viewbox is associated, the pointer returned will be a null
       * smart pointer.
       */
      Viewbox::pointer viewbox();

      /**
       * Sets this Drawable's viewbox to the provided viewbox.
       *
       * This method does not copy the viewbox, but will hold shared ownership
       * of the shared smart pointer.
       *
       * Calling set_viewbox() with no parameters sets the viewbox to a
       * null pointer, thus removing it.
       */
      void set_viewbox(Viewbox::pointer viewbox = Viewbox::pointer() );

      /**
       * Sets this Drawable's viewbox to the provided viewbox.
       *
       * This method will not maintain a connection to the provided Viewbox,
       * and will create its own copy.
       */
      void set_viewbox(const Viewbox& viewbox );

      /** Signal emitted with x,y,w,h values when this Drawable needs to be redrawn. */
      sigc::signal<void, double, double, double, double> signal_redraw();

      /** Signal emitted when the xy position is changed */
      sigc::signal<void> signal_xy();

      /** Signal emitted when the scale is changed */
      sigc::signal<void> signal_scale();

      /** Signal emitted when the rotation value is changed */
      sigc::signal<void> signal_rotation();

      /** Signal emitted when the skew is changed */
      sigc::signal<void> signal_skew();

      /** Signal emitted when the selectable value is changed */
      sigc::signal<void> signal_selectable();

      /** Signal emitted when the centroid is changed */
      sigc::signal<void> signal_centroid();

      /**
       * This is a proxy for the virtual need_redraw() method. The changed signal of the
       * viewbox is connected to this proxy that in turn calls the appropriate version
       * of need_redraw().
       * HACK
       * For some reason child Shape can't access this parent method when protected
       */
      void need_redraw_proxy();

      virtual Drawable::pointer clone() const = 0;

      virtual Group* parent();

      PAPYRUS_CLASS_NAME("Drawable");

      protected:
      Group* m_parent;

      Group* m_drawing_parent;

      bool m_visible;

      /** x translation value */
      double m_tx;

      /** y translation value */
      double m_ty;

      /** x scaling value */
      double m_sx;

      /** y scaling value */
      double m_sy;

      /** rotation value in radians */
      double m_r;

      /** skew x angle in radians */
      double m_skewx;

      /** skew y angle in radians */
      double m_skewy;

      /** whether the drawable is selectable */
      bool m_selectable;

      /** x centroid value */
      double m_centroid_x;

      /** y centroid value */
      double m_centroid_y;

      bool m_exclude_from_extents;

      /**
       * Pointer to a viewbox, which may be null if this Drawable doesn't have a viewbox.
       */
      Viewbox::pointer m_viewbox;

      sigc::connection m_viewbox_connection;

      /**
       * True if the cairo matrix needs to be regenerated before rendering or
       * returning from the matrix method.
       */
      bool m_regenerate_matrix;

      /** The cairo matrix returned from matrix and used in rendering. */
      Cairo::Matrix m_matrix;

      /**
       * Recalculates m_matrix if m_regenerate_matrix is true and sets
       * m_regenerate_matrix to false; does nothing if m_regenerate_matrix
       * is false.
       */
      void recalculate_matrix();

      /**
       * Signal emitted when this object has changed and needs to be redrawn.
       */
      sigc::signal<void, double, double, double, double> m_signal_redraw;

      /** Signal emitted when the xy position is changed */
      sigc::signal<void> m_signal_xy;

      /** Signal emitted when the scale is changed */
      sigc::signal<void> m_signal_scale;

      /** Signal emitted when the rotation value is changed */
      sigc::signal<void> m_signal_rotation;

      /** Signal emitted when the skew is changed */
      sigc::signal<void> m_signal_skew;

      /** Signal emitted when the selectable value is changed */
      sigc::signal<void> m_signal_selectable;

      /** Signal emitted when the centroid is changed */
      sigc::signal<void> m_signal_centroid;

      friend class Group;

      /**
       * Convenience method that children can call and will take care of emitting
       * the need redraw signal with the proper parameters.
       *
       * For efficiency, if a child needs less than a complete redraw the child
       * could reimplement this method.
       */
      virtual void need_redraw( );

      /**
       * To be used by children that need to set the extents.
       * Direct access is not allowed to ensure that children go through
       * this method, which may be polymorphically implemented by children
       * such as Shape.
       */
      virtual void set_extents( const Region& extents );

    private:
      /**
       * The private extents. Children should use the set_extents()
       * method when the extents have changed.
       */
      Region m_extents, m_extentsold;

    };

}

#endif
