/***************************************************************************
 *   Copyright (C) 2004 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#include "group.h"

#include <assert.h>

#include <iostream>
#include <sstream>
#include <algorithm>

namespace Papyrus
{

  Group::Group():m_size(0)
  {
    set_selectable( false );
  }

  Group::pointer Group::create()
  {
    return Group::pointer( new Group() );
  }

  Group::~Group()
  {
    Layers::iterator ilayer;
    Layer::iterator ichild;
    Connections::iterator conniter;

    for ( conniter = m_redraw_connections.begin(); conniter != m_redraw_connections.end(); conniter++ )
      conniter->second.disconnect();

    for ( ilayer = m_layers.begin(); ilayer != m_layers.end(); ilayer++ )
      for ( ichild = ilayer->second.begin(); ichild != ilayer->second.end(); ichild++ )
        ( *ichild ) ->m_parent = NULL;
  }

  bool Group::add( Drawable::pointer item, int layer, bool drawing_only )
  {
    Connections::iterator i;

    // We must have a valid pointer
    if ( ! item ) return false;

    // we can't add an item to itself, otherwise we'll end up with recursive problems
    if ( item.get() == this ) return false;

    if ( drawing_only ) {

      // if we're already the drawing parent, we're done
      if ( item->m_drawing_parent == this )
        return true;

      // if it already has a drawing parent, let's disconnect it from need_redraw messages
      if ( item->m_drawing_parent ) {
        i = m_redraw_connections.find( item );
        if ( i != m_redraw_connections.end() ) {
          i->second.disconnect();
          m_redraw_connections.erase( i );
        }
      }
      // this group becomes the new drawing parent
      item->m_drawing_parent = this;

      // we need the redraw signal connected
      // we'll connect the changed signal later
      m_redraw_connections[ item ] =
          item->signal_redraw().connect( sigc::bind( sigc::mem_fun( *this, &Group::on_child_need_redraw ), item ) );

    } else { // not drawing only, but a normal add

      // if we're already the parent, we're done
      if ( item->m_parent == this )
        return true;

      // if it already has a parent, we need to remove it from the old parent
      if ( item->m_parent != NULL )
        item->m_parent->remove( item );

      // now, let's make this group the parent
      item->m_parent = this;

      // if it doesn't already have a drawing parent, we'll make this the drawing parent
      if ( item->m_drawing_parent == NULL ) {
        item->m_drawing_parent = this;

        // we need the redraw signal connected
        // we'll connect the changed signal later
        m_redraw_connections[ item ] =
            item->signal_redraw().connect( sigc::bind( sigc::mem_fun( *this, &Group::on_child_need_redraw ), item ) );
      }
    }

    // Add this object to the list of children
    m_layers[layer].push_front( item );

    m_size++;

    // Connect to the changed signal
    m_changed_connections[ item ] =
        item->signal_changed().connect( sigc::bind( sigc::mem_fun( *this, &Group::on_child_changed ), item ) );

    // Include the child's extents in this group's extents
    m_extents.include( item->global_extents() );

    // Perform any work needed to update the extents
    this->update_extents();

    // Notify anyone connected to the child_added signal
    m_signal_child_added.emit( item );

    // We'll need a redraw of course
    this->need_redraw();

    return true;
  }

  bool Group::remove( Drawable::pointer item )
  {
    bool removed = false;
    Layers::iterator ilayer;
    Layer::iterator ichild;
    Connections::iterator i;

    // We must have a valid pointer
    if ( ! item )
      return false;

    for (ilayer = m_layers.begin(); ilayer != m_layers.end(); ilayer++)
    {
      ichild = std::find( ilayer->second.begin(), ilayer->second.end(), item );
      if ( ichild != ilayer->second.end() )
      {
        ilayer->second.erase( ichild );
        m_size--;
        if ( ilayer->second.size() == 0 )
          m_layers.erase(ilayer);
        removed = true;
        break;
      }
    }

    if ( !removed )
      return false;

    // Let's find out if we have a redraw signal for it
    i = m_redraw_connections.find( item );
    if ( i != m_redraw_connections.end() ) {
      // if we do, let's disconnect and remove the connection
      i->second.disconnect();
      m_redraw_connections.erase( i );
    }

    // Let's find out if we have a changed signal connection
    i = m_changed_connections.find( item );
    if ( i != m_changed_connections.end() ) {
      // if we do, let's disconnect and remove the connection
      i->second.disconnect();
      m_changed_connections.erase( i );
      removed = true;
    }

    // find out if we were the drawing parent only
    // if we were, let's make the real parent the drawing parent
    if ( item->m_drawing_parent == this && item->m_parent != this ) {
      item->m_drawing_parent = item->m_parent;
      if ( item->m_drawing_parent ) {
        item->m_drawing_parent->m_redraw_connections[ item ] =
            item->signal_redraw().connect( sigc::bind( sigc::mem_fun( *(item->m_drawing_parent), &Group::on_child_need_redraw ), item )  );
        item->m_drawing_parent->need_redraw();
      }
    }
    // if we weren't the drawing parent, let's see if we're the regular parent
    else if ( item->m_parent == this )
      item->m_parent = NULL;

    this->update_extents();
    m_signal_child_removed.emit( item );
    this->need_redraw();

    return true;
  }

  bool Group::remove( Drawable* object )
  {
    Layers::iterator ilayer;
    Layer::iterator ichild;

    // We must have a valid pointer
    if ( object == NULL )
      return false;

    // Let's look for the pointer and then call the remove method with the smart pointer
    for ( ilayer = m_layers.begin(); ilayer != m_layers.end(); ilayer++ ) {
      for ( ichild = ilayer->second.begin(); ichild != ilayer->second.end(); ichild++ ) {
        if ( ichild->get() == object )
          return this->remove( *ichild );
      }
    }
    return false;
  }

  bool Group::clear()
  {
    Layers::iterator ilayer;

    while ( m_layers.size() > 0 )
      remove( m_layers.begin()->second.front() );

    update_extents();
    this->need_redraw();
    return true;
  }

  void Group::draw( Cairo::RefPtr<Cairo::Context> cairo )
  {
    Layers::iterator ilayer;
    Layer::reverse_iterator ichild;

    for ( ilayer = m_layers.begin(); ilayer != m_layers.end(); ilayer++ )
    {
      for ( ichild = ilayer->second.rbegin(); ichild != ilayer->second.rend(); ichild++ )
      {
        if ( (*ichild)->m_drawing_parent == this )
          ( *ichild ) ->render( cairo );
      }
    }
  }

  void Group::on_child_need_redraw( double x, double y, double w, double h, Drawable::pointer child )
  {
    Region region( x, y, w, h );
    recalculate_matrix();
    region.transform( m_matrix );
    region.xywh( x, y, w, h );
    m_signal_redraw.emit( x, y, w, h );
  }

  bool Group::raise( Drawable::pointer item, int steps )
  {
    Layers::iterator ilayer;
    Layer::iterator ichild1, ichild2;

    if ( steps < 0 ) return this->lower(item, -steps);
    if ( steps == 0 ) return true;

    for ( ilayer = m_layers.begin(); ilayer != m_layers.end(); ilayer++ )
    {
      ichild2 = std::find( ilayer->second.begin(), ilayer->second.end(), item );
      if ( ichild2 != ilayer->second.end() )
      {
        ichild1 = ichild2;
        for (int i=0; i<steps && ichild1 != ilayer->second.begin(); i++)
          ichild1--;
        ilayer->second.erase( ichild2 );
        ilayer->second.insert( ichild1, item );
        item->need_redraw();
        return true;
      }
    }
    return false;
  }

  bool Group::raise_to_top( Drawable::pointer item )
  {
    Layers::iterator ilayer;
    Layer::iterator ichild;

    for ( ilayer = m_layers.begin(); ilayer != m_layers.end(); ilayer++ )
    {
      ichild = std::find( ilayer->second.begin(), ilayer->second.end(), item );
      if ( ichild != ilayer->second.end() && ichild != ilayer->second.begin() )
      {
        ilayer->second.erase( ichild );
        ilayer->second.push_front( item );
        item->need_redraw();
        return true;
      }
    }
    return false;
  }

  bool Group::lower( Drawable::pointer item, int steps )
  {
    Layers::iterator ilayer;
    Layer::iterator ichild1, ichild2;

    if ( steps < 0 ) return this->lower(item, -steps);
    if ( steps == 0 ) return true;

    for ( ilayer = m_layers.begin(); ilayer != m_layers.end(); ilayer++ )
    {
      ichild2 = std::find( ilayer->second.begin(), ilayer->second.end(), item );
      if ( ichild2 != ilayer->second.end() )
      {
        ichild1 = ichild2;
        for (int i=0; i<steps+1 && ichild1 != ilayer->second.end(); i++)
          ichild1++;
        ilayer->second.erase( ichild2 );
        ilayer->second.insert( ichild1, item );
        item->need_redraw();
        return true;
      }
    }
    return false;
  }

  bool Group::lower_to_bottom( Drawable::pointer item )
  {
    Layers::iterator ilayer;
    Layer::iterator ichild;

    for ( ilayer = m_layers.begin(); ilayer != m_layers.end(); ilayer++ )
    {
      ichild = std::find( ilayer->second.begin(), ilayer->second.end(), item );
      // is it in this layer?
      if ( ichild != ilayer->second.end() )
      {
        if ( ++ichild == ilayer->second.end() )
        {
          // it's already at the bottom... nothing to do
          return true;
        }
        else
        {
          --ichild; // decrement the iterator back to the item
          ilayer->second.erase( Layer::iterator(ichild) );
          ilayer->second.push_back( item );
          item->need_redraw();
          return true;
        }
      }
    }
    return false;
  }

  bool Group::move_to_layer( int layer, Drawable::pointer item )
  {
    Layer::iterator ichild;

    for ( Layers::iterator ilayer = m_layers.begin(); ilayer != m_layers.end(); ilayer++ )
    {
      ichild = std::find( ilayer->second.begin(), ilayer->second.end(), item );
      if ( ichild != ilayer->second.end() )
      {
        if ( ilayer->first == layer ) return true;
        ilayer->second.erase( ichild );
        if ( ilayer->first > layer )
          m_layers[layer].push_front( item );
        else
          m_layers[layer].push_back( item );
        this->need_redraw();
        return true;
      }
    }

    return false;
  }

  bool Group::inside( double x, double y )
  {
    Layers::iterator ilayer;
    for ( ilayer = m_layers.begin(); ilayer != m_layers.end(); ilayer++ )
    {
      for ( Layer::iterator ichild = ilayer->second.begin(); ichild != ilayer->second.end(); ichild++ )
        if ( ( *ichild ) ->inside( x, y ) )
          return true;
    }
    return false;
  }

  sigc::signal<void, Drawable::pointer>& Group::signal_child_added()
  {
    return m_signal_child_added;
  }

  sigc::signal<void, Drawable::pointer>& Group::signal_child_removed()
  {
    return m_signal_child_removed;
  }

  bool Group::is_group()
  {
    return true;
  }

  std::vector<Drawable::pointer> Group::select( double x, double y, unsigned depth )
  {
    Layers::iterator ilayer;
    std::vector<Drawable::pointer> result, intermediate;
    Group::pointer g;

    if ( depth == 0 )
      return result;

    // transform the points according to the group
    // affine matrix and check each of the children
    cairo_matrix_t inverse = matrix();
    cairo_matrix_invert( &inverse );
    cairo_matrix_transform_point( &inverse, &x, &y );
    for ( ilayer = m_layers.begin(); ilayer != m_layers.end(); ilayer++ )
    {
      for ( Layer::iterator ichild = ilayer->second.begin(); ichild != ilayer->second.end(); ichild++ ) {
        if ( ( *ichild ) ->inside( x, y ) )
        {
          if ( ( *ichild ) ->is_group() )
          {
            g = papyrus_smart_pointer_cast<Group>( ( *ichild ) );
            if ( g->is_selectable() )
            {
              result.push_back( g );
              depth--;
            }
            intermediate = g->select( x, y, depth );
            result.insert( result.end(), intermediate.begin(), intermediate.end() );
          }
          else
          {
            result.push_back( *ichild );
            depth--;
          }
        }
        if ( depth == 0 )
          break;
      }
    }
    return result;
  }

  void Group::rebuild_extents()
  {
    Layers::iterator ilayer;
    bool first_child = true;

    // We'll start with null extents. But, if we find at least one child
    // we'll start building our extents from that child's extents so that
    // we don't end up including the point (0,0) in the extents.
    m_extents.set_xywh( 0.0, 0.0, 0.0, 0.0 );

    for ( ilayer = m_layers.begin(); ilayer != m_layers.end(); ilayer++ )
    {
      for ( Layer::iterator ichild = ilayer->second.begin(); ichild != ilayer->second.end(); ichild++ )
      {
        if ( (*ichild)->exclude_from_extents() || ! (*ichild)->is_visible() )
          continue;
        if ( first_child )
        {
          m_extents = (*ichild)->global_extents();
          first_child = false;
        }
        else
          m_extents.include( (*ichild)->global_extents() );
      }
    }
  }

  void Group::on_child_changed( Drawable::pointer )
  {
    //   if (changed & RENDER_BBOX)
    update_extents();
  }

  size_t Group::size() const
  {
    return m_size;
  }

  const Group::Layers& Group::layers( ) const
  {
    return m_layers;
  }

  const Group::Layer& Group::layer( int l ) const throw (std::out_of_range)
  {
    Layers::const_iterator ilayer = m_layers.find(l);
    if ( ilayer == m_layers.end() ) {
      std::ostringstream sout;
      sout << "Papyrus::Group::layer() Group does not have requested layer "
           << l;
      throw std::out_of_range(sout.str());
    }
    return ilayer->second;
  }

  bool Group::has_layer( int l )
  {
    if ( m_layers.find(l) == m_layers.end() ) return false;
    return true;
  }

  void Group::update_extents( )
  {
    rebuild_extents();
    this->need_redraw();
  }

}
