/***************************************************************************
 *   Copyright (C) 2004 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   Copyright (C) 2007 Peter Miller                                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#include "image.h"

#include <byteswap.h>

#include <endian.h>

namespace Papyrus
{

  Image::Image( ) :
      Drawable(),
      m_width ( 0.0 ),
      m_height ( 0.0 ) {}

  Image::Image ( std::string filename )  :
      Drawable(),
      m_width ( 0.0 ),
      m_height ( 0.0 )
  {
    this->set_image ( filename );
  }

  Image::Image ( unsigned char* data, Cairo::Format format, int width, int height, int stride, bool convert_rgba )
  {
    this->set_image ( data, format, width, height, stride, convert_rgba );
  }

  Image::Image ( Cairo::Format format, int width, int height )
  {
    this->set_image ( format, width, height );
  }

  Image::Image ( Cairo::RefPtr<Cairo::ImageSurface> img )
  {
    this->set_image ( img );
  }

  Image::pointer Image::create( )
  {
    return Image::pointer ( new Image( ) );
  }

  Image::pointer Image::create ( std::string filename )
  {
    return Image::pointer ( new Image ( filename ) );
  }

  Image::pointer Image::create ( unsigned char* data, Cairo::Format format, int width, int height, int stride, bool convert_rgba )
  {
    return Image::pointer ( new Image ( data, format, width, height, stride, convert_rgba ) );
  }

  Image::pointer Image::create ( Cairo::Format format, int width, int height )
  {
    return Image::pointer ( new Image ( format, width, height ) );
  }

  Image::pointer Image::create ( Cairo::RefPtr<Cairo::ImageSurface> img )
  {
    return pointer ( new Image ( img ) );
  }

  Image::~Image() {}

  void Image::draw ( Cairo::RefPtr<Cairo::Context> cairo )
  {
    if ( m_image )
    {
      cairo->set_source ( m_image, 0, 0 );
      cairo->paint( );
    }
  }

  std::string Image::filename( ) const
  {
    return m_filename;
  }

  void Image::set_image ( const std::string & filename )
  {
    m_filename = filename;
    m_image.clear();
    if ( m_filename.size() > 0 )
      m_image = Cairo::ImageSurface::create_from_png ( m_filename );
    this->on_image_changed();
  }

  void Image::set_image ( unsigned char* data, Cairo::Format format, int width, int height, int stride, bool convert_rgba )
  {
    uint32_t* data32;
    uint32_t alpha;
    m_filename.clear();
    m_image.clear();

    if ( convert_rgba )
    {
      data32 = ( uint32_t* ) ( data );
      for ( int i = 0; i < width*height; i++ )
      {
#ifdef __LITTLE_ENDIAN
        data32[i] = bswap_32 ( data32[i] );
#endif
        alpha = data32[i] << 24;
        data32[i] >>= 8;
        data32[i] |= alpha;
      }
    }

    m_image = Cairo::ImageSurface::create ( data, format, width, height, stride );
    this->on_image_changed();
  }

  void Image::set_image ( Cairo::Format format, int width, int height )
  {
    m_filename.clear();
    m_image.clear();
    m_image = Cairo::ImageSurface::create ( format, width, height );
    this->on_image_changed();
  }

  void Image::set_image ( Cairo::RefPtr<Cairo::ImageSurface> img )
  {
    m_filename.clear();
    m_image = img;
    this->on_image_changed();
  }

  double Image::width()
  {
    return m_width;
  }

  double Image::height()
  {
    return m_height;
  }

  void Image::width_height ( double& width, double& height )
  {
    width = m_width;
    height = m_height;
  }

  void Image::on_image_changed()
  {
    if ( ! m_image )
    {
      m_width = 0.0;
      m_height = 0.0;
    }
    else
    {
      m_width = m_image->get_width( );
      m_height = m_image->get_height( );
      m_centroid_x = m_width / 2.0;
      m_centroid_y = m_height / 2.0;
    }

    this->set_extents ( Region ( 0.0, 0.0, m_width, m_height ) );
    this->need_redraw();
  }

}
