/***************************************************************************
 *   Copyright (C) 2004 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#include "linestyle.h"

namespace Papyrus
{

  LineStyle::LineStyle( Cairo::RefPtr<Cairo::Pattern> pattern, double width ) :
      m_pattern( pattern ),
      m_width( width ),
      m_cap( Cairo::LINE_CAP_BUTT ),
      m_join( Cairo::LINE_JOIN_MITER ),
      m_miter_limit( INFINITY ),
      m_dash_offset( 0.0 )
  {}

  LineStyle::LineStyle( double width, Cairo::LineCap cap, Cairo::LineJoin join ) :
      m_width( width ),
      m_cap( cap ),
      m_join( join ),
      m_miter_limit( INFINITY ),
      m_dash_offset( 0.0 )
  { }

  LineStyle::LineStyle( Cairo::RefPtr<Cairo::Pattern> pattern,
                        double width, Cairo::LineCap cap, Cairo::LineJoin join ) :
      m_pattern( pattern ),
      m_width( width ),
      m_cap( cap ),
      m_join( join ),
      m_miter_limit( INFINITY ),
      m_dash_offset( 0.0 )
  { }

  LineStyle::~LineStyle()
  { }

  LineStyle & LineStyle::operator =( const LineStyle & other )
  {
    m_pattern = other.m_pattern;
    m_width = other.m_width;
    m_cap = other.m_cap;
    m_join = other.m_join;
    m_miter_limit = other.m_miter_limit;
    m_dash_array = other.m_dash_array;
    m_dash_offset = other.m_dash_offset;

    m_signal_pattern.emit();
    m_signal_width.emit();
    m_signal_cap.emit();
    m_signal_join.emit();
    m_signal_miter_limit.emit();
    m_signal_dash.emit();
  }

  bool LineStyle::operator==( const LineStyle& other ) const
  {
    // TODO this is kind of hacked and needs to be improved
    return ( m_pattern == other.m_pattern &&
             m_width == other.m_width &&
             m_cap == other.m_cap &&
             m_join == other.m_join &&
             m_miter_limit == other.m_miter_limit &&
//              m_dash_array == other.m_dash_array &&
             m_dash_offset == other.m_dash_offset );
  }

  bool LineStyle::operator!=( const LineStyle& other ) const
  {
    return ! this->operator==(other);
  }

  const std::valarray< double >& LineStyle::dash_array()
  {
    return m_dash_array;
  }

  double LineStyle::dash_offset()
  {
    return m_dash_offset;
  }

  void LineStyle::set_dash ( const std::valarray< double > dashes, double offset )
  {
    m_dash_array = dashes;
    m_dash_offset = offset;
    m_signal_dash.emit();
    m_signal_changed.emit();
  }

  Cairo::LineCap LineStyle::cap()
  {
    return m_cap;
  }

  void LineStyle::set_cap ( Cairo::LineCap cap )
  {
    m_cap = cap;
    m_signal_cap.emit();
    m_signal_changed.emit();
  }

  Cairo::LineJoin LineStyle::join()
  {
    return m_join;
  }

  void LineStyle::set_join ( Cairo::LineJoin join )
  {
    m_join = join;
    m_signal_join.emit();
    m_signal_changed.emit();
  }

  double LineStyle::width()
  {
    return m_width;
  }

  void LineStyle::set_width ( double width )
  {
    m_width = width;
    m_signal_width.emit();
    m_signal_changed.emit();
  }

  double LineStyle::miter_limit()
  {
    return m_miter_limit;
  }

  void LineStyle::set_miter_limit ( double limit )
  {
    m_miter_limit = limit;
    m_signal_miter_limit.emit();
    m_signal_changed.emit();
  }

  Cairo::RefPtr<Cairo::Pattern> LineStyle::pattern()
  {
    return m_pattern;
  }

  void LineStyle::set_pattern( Cairo::RefPtr<Cairo::Pattern> pattern )
  {
    m_pattern = pattern;
    m_signal_pattern.emit();
    m_signal_changed.emit();
  }

  void LineStyle::render( Cairo::RefPtr<Cairo::Context> cairo )
  {
    if ( m_pattern )
      cairo->set_source( m_pattern );
    cairo->set_line_width( m_width );
    cairo->set_line_cap( m_cap );
    cairo->set_line_join( m_join );
    if ( m_join == Cairo::LINE_JOIN_MITER )
      cairo->set_miter_limit( m_miter_limit );
    if ( m_dash_array.size() > 0 )
      cairo->set_dash( m_dash_array, m_dash_offset );
  }

  sigc::signal<void> LineStyle::signal_pattern()
  {
    return m_signal_pattern;
  }

  sigc::signal<void> LineStyle::signal_width()
  {
    return m_signal_width;
  }

  sigc::signal<void> LineStyle::signal_cap()
  {
    return m_signal_cap;
  }

  sigc::signal<void> LineStyle::signal_join()
  {
    return m_signal_join;
  }

  sigc::signal<void> LineStyle::signal_miter_limit()
  {
    return m_signal_miter_limit;
  }

  sigc::signal<void> LineStyle::signal_dash()
  {
    return m_signal_dash;
  }

}
