/***************************************************************************
 *   Copyright (C) 2004 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#include "scaler.h"

namespace Papyrus
{

  Scaler::Scaler( Group::pointer selection_group,
                          SelectionMode mode,
                          int selection_depth,
                          unsigned sides )
      : Selector( selection_group, mode, selection_depth ),
        m_sides( sides )
  {
    m_original_position[ 0 ] = 0.0;
    m_original_position[ 1 ] = 0.0;
    m_grid[ 0 ] = 0.0;
    m_grid[ 1 ] = 0.0;
  }

  Scaler::Scaler( Group::pointer select_from_group,
                          Group::pointer selection_group,
                          SelectionMode mode,
                          int selection_depth,
                          unsigned sides )
      : Selector( select_from_group, mode, selection_depth ),
        m_sides( sides )
  {
    this->set_selection_group(selection_group);
    m_original_position[ 0 ] = 0.0;
    m_original_position[ 1 ] = 0.0;
    m_grid[ 0 ] = 0.0;
    m_grid[ 1 ] = 0.0;
  }

  Scaler::pointer Scaler::create( Group::pointer selection_group,
                                          SelectionMode mode,
                                          int selection_depth,
                                          unsigned sides )
  {
    return Scaler::pointer( new Scaler( selection_group, mode, selection_depth, sides ) );
  }


  Scaler::pointer Scaler::create( Group::pointer select_from_group,
                                          Group::pointer selection_group,
                                          SelectionMode mode,
                                          int selection_depth,
                                          unsigned sides )
  {
    return Scaler::pointer( new Scaler( select_from_group, selection_group, mode, selection_depth, sides ) );
  }


  Scaler::~Scaler()
  {}

  void Scaler::set_selection_group( Group::pointer selection_group, bool add_drawable_only )
  {
    Selector::set_selection_group( selection_group, add_drawable_only );

    if ( selection_group->class_name() == "Handle Box" ) {
      m_handlebox = papyrus_smart_pointer_cast<Handlebox>(selection_group);
    } else {
      m_handlebox = Handlebox::pointer();
    }
  }

  unsigned Scaler::sides()
  {
    return m_sides;
  }

  void Scaler::set_sides( unsigned sides )
  {
    m_sides = sides;
  }

  void Scaler::enable_sides( unsigned sides )
  {
    m_sides |= sides;
  }

  void Scaler::disable_sides( unsigned sides )
  {
    m_sides &= ~sides;
  }

//   void Scaler::grid( double & x, double & y )
//   {
//     x = m_grid[ 0 ];
//     y = m_grid[ 1 ];
//   }
//
//   void Scaler::set_grid( double x, double y )
//   {
//     m_grid[ 0 ] = x;
//     m_grid[ 1 ] = y;
//   }
//
//   void Scaler::set_grid_x( double x )
//   {
//     m_grid[ 0 ] = x;
//   }
//
//   void Scaler::set_grid_y( double y )
//   {
//     m_grid[ 1 ] = y;
//   }

  bool Scaler::on_button_press( const Event::ButtonPress& event )
  {
    int h = -1;

    if ( m_handlebox ) {
      h = m_handlebox->inside_handle( event.x, event.y );

      switch (h) {
        case 0:
          this->set_sides(TOP);
          break;
        case 1:
          this->set_sides(TOP|RIGHT);
          break;
        case 2:
          this->set_sides(RIGHT);
          break;
        case 3:
          this->set_sides(BOTTOM|RIGHT);
          break;
        case 4:
          this->set_sides(BOTTOM);
          break;
        case 5:
          this->set_sides(BOTTOM|LEFT);
          break;
        case 6:
          this->set_sides(LEFT);
          break;
        case 7:
          this->set_sides(TOP|LEFT);
          break;
        default:
          Selector::on_button_press( event );
          this->set_sides(0);
          break;
      }
    }
    else
      Selector::on_button_press( event );

    m_original_position[ 0 ] = event.x;
    m_original_position[ 1 ] = event.y;
    m_last_position[ 0 ] = event.x;
    m_last_position[ 1 ] = event.y;

    return true;
  }

  bool Scaler::on_motion( const Event::Motion & event )
  {
    Selection::iterator i;
    double tx = 0.0, ty = 0.0, sx=0.0, sy=0.0;
    double dx, dy, ldx, ldy;
    Region extents, pre_scale_extents, post_scale_extents;

    // calculate the movement
    dx = event.x - m_last_position[0];
    dy = event.y - m_last_position[1];

    // calculate the contained global extents
    extents = this->extents();

//     if ( m_sides & TRANSLATE_X ) {
//       tx = event.x - m_last_position[ 0 ];
//       if ( m_grid[ 0 ] > 0.0 )
//         tx = round( tx / m_grid[ 0 ] ) * m_grid[ 0 ];
//     }
//
//     if ( m_sides & TRANSLATE_Y ) {
//       ty = event.y - m_last_position[ 1 ];
//       if ( m_grid[ 1 ] > 0.0 )
//         ty = round( ty / m_grid[ 1 ] ) * m_grid[ 1 ];
//     }
//

    if ( m_sides & RIGHT ) dx = -dx;
    if ( m_sides & ( LEFT | RIGHT ) ) sx = (extents.width() - dx) / extents.width();
    if ( m_sides & BOTTOM ) dy = -dy;
    if ( m_sides & ( TOP | BOTTOM ) ) sy = (extents.height() - dy) / extents.height();

    for ( i = m_selection.begin(); i != m_selection.end(); i++ ) {
      pre_scale_extents = (*i)->local_extents();

      if ( sx != 0.0 && sy != 0.0 ) ( *i ) ->scale( sx, sy );
      else if ( sx != 0.0 )         ( *i ) ->scale_x( sx );
      else if ( sy != 0.0 )         ( *i ) ->scale_y( sy );

      post_scale_extents = (*i)->local_extents();

      if ( m_sides & LEFT && ! (m_sides & RIGHT) )
        tx = pre_scale_extents.right() - post_scale_extents.right();
      else if ( ! (m_sides & LEFT) && m_sides & RIGHT )
        tx = pre_scale_extents.left() - post_scale_extents.left();
      else if ( m_sides & LEFT && m_sides & RIGHT )
        tx = ( pre_scale_extents.left() - post_scale_extents.left() + pre_scale_extents.right() - post_scale_extents.right() ) / 2.0;

      if ( m_sides & TOP && ! (m_sides & BOTTOM) )
        ty = pre_scale_extents.bottom() - post_scale_extents.bottom();
      else if ( ! (m_sides & TOP) && m_sides & BOTTOM )
        ty = pre_scale_extents.top() - post_scale_extents.top();
      else if ( m_sides & TOP && m_sides & BOTTOM )
        ty = ( pre_scale_extents.top() - post_scale_extents.top() + pre_scale_extents.bottom() - post_scale_extents.bottom() ) / 2.0;

      if ( tx != 0.0 && ty != 0.0 )
        ( *i ) ->translate( tx, ty );
      else if ( tx != 0.0 )
        ( *i ) ->translate_x( tx );
      else if ( ty != 0.0 )
        ( *i ) ->translate_y( ty );

    }

    extents = this->extents();

    m_last_position[ 0 ] = event.x;
    m_last_position[ 1 ] = event.y;

    return true;
  }

}
