/***************************************************************************
 *   Copyright (C) 2004 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#include "shape.h"

namespace Papyrus
{

  Shape::Shape( Cairo::RefPtr<Cairo::Pattern> fill )
      : Drawable(),
      m_fill( fill ),
      m_use_outline( false ),
  m_fill_path(NULL),
  m_outline_path(NULL)
  {
    m_outline_changed_connection = m_outline.signal_changed().connect( sigc::mem_fun( *this, &Shape::on_outline_changed ) );
    m_outline_changed_connection.block();
  }

  Shape::Shape( Cairo::RefPtr<Cairo::Pattern> fill,
                const LineStyle& outline )
      : Drawable(),
      m_fill( fill ),
      m_outline( outline ),
      m_use_outline( true ),
  m_fill_path(NULL),
  m_outline_path(NULL)
  {
    m_outline_changed_connection = m_outline.signal_changed().connect( sigc::mem_fun( *this, &Shape::on_outline_changed ) );
    m_outline_changed_connection.block();
  }


  Shape::~Shape() {
    if ( m_fill_path )
      free( m_fill_path );
    if ( m_outline_path )
      free( m_outline_path );
  }


  void Shape::draw( Cairo::RefPtr<Cairo::Context> cairo )
  {
    if ( m_fill )
    {
      if ( !m_fill_path )
      {
        this->draw_shape( cairo );
        m_fill_path = cairo->copy_path();
      }
      else
        cairo->append_path( *m_fill_path );

      cairo->set_source( m_fill );
      cairo->fill( );
    }

    if ( m_use_outline )
    {
      m_outline.render( cairo );

      if ( !m_outline_path )
      {
        this->draw_outline( cairo );
        m_outline_path = cairo->copy_path();
      }
      else
        cairo->append_path( *m_outline_path );

      cairo->stroke( );
    }
  }

  void Shape::draw_outline( Cairo::RefPtr< Cairo::Context > cairo )
  {
    this->draw_shape( cairo );
  }

  Cairo::RefPtr< Cairo::Pattern > Shape::fill( )
  {
    return m_fill;
  }

  void Shape::set_fill( Cairo::RefPtr<Cairo::Pattern> fill )
  {
    m_fill = fill;
    this->need_redraw();
  }

  LineStyle& Shape::outline()
  {
    return m_outline;
  }

  void Shape::set_outline( const LineStyle& outline, bool use_outline )
  {
    m_use_outline = use_outline;
    m_outline = outline;
    if ( m_use_outline )
      m_outline_changed_connection.unblock();
    else
      m_outline_changed_connection.block();
    this->on_outline_changed();
  }

  void Shape::set_outline( Cairo::RefPtr<Cairo::Pattern> pattern, bool use_outline )
  {
    m_use_outline = use_outline;
    m_outline.set_pattern( pattern );
    if ( m_use_outline )
      m_outline_changed_connection.unblock();
    else
      m_outline_changed_connection.block();
    this->on_outline_changed();
  }

  bool Shape::use_outline()
  {
    return m_use_outline;
  }

  void Shape::set_use_outline( bool use_outline )
  {
    m_use_outline = use_outline;
    if ( m_use_outline )
      m_outline_changed_connection.unblock();
    else
      m_outline_changed_connection.block();
    this->on_outline_changed();
  }

  void Shape::set_extents( const Region & extents )
  {
    m_internal_extents = extents;
    if ( m_use_outline )
    {
      double x, y, w, h, offset;
      Region temp_region = extents;
      extents.xywh( x, y, w, h );
      offset = m_outline.width();
      x -= offset;
      y -= offset;
      w += 2.0 * offset;
      h += 2.0 * offset;
      temp_region.set_xywh( x, y, w, h );
      Drawable::set_extents( temp_region );
    }
    else
      Drawable::set_extents( extents );
  }

  void Shape::on_outline_changed()
  {
    this->set_extents( m_internal_extents );
    this->need_redraw();
  }

  void Shape::shape_changed( unsigned which )
  {
    if ( which & FILL ) {
      free( m_fill_path );
      m_fill_path = NULL;
    }
    if ( which & OUTLINE ) {
      free( m_outline_path );
      m_outline_path = NULL;
    }
    this->need_redraw();
  }

}
