/***************************************************************************
 *   Copyright (C) 2004 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#ifndef PAPYRUSSHAPE_H
#define PAPYRUSSHAPE_H

#include <papyrus/enums.h>
#include <papyrus/drawable.h>
#include <papyrus/linestyle.h>

namespace Papyrus
{

  /**
   * Papyrus::Shape inherits the concepts of spatial placement and meaningful affine transforms
   * from its parent Papyrus::Drawable, and extends it with the concept of outlining and filling.
   *
   * When inheriting from Papyrus::Shape you should provide a child specific implementation
   * of the draw_shape() method (and not the draw() method from Papyrus::Drawable). This method
   * should use cairo methods to draw the structure of your shape that will be filled with the
   * pattern set by calling the set_fill() method.
   *
   * In most cases, providing a child specific implementation of the draw_shape() method will
   * be sufficient. However, if your child needs specific methods to perform outlining you
   * should also override the draw_outline() method.
   *
   * @author Rick L Vinyard Jr
   */
  class Shape : public Drawable
  {
  protected:

      /** Constructor that accepts a fill pattern */
      Shape( Cairo::RefPtr<Cairo::Pattern> fill = Cairo::RefPtr<Cairo::Pattern>() );

      /** Constructor that accepts a fill and outline pattern */
      Shape( Cairo::RefPtr<Cairo::Pattern> fill, const LineStyle& outline );

  public:
    typedef PapyrusSmartPointer<Shape> pointer;

    /** Destructor */
    virtual ~Shape();

    /**
     * Override of Drawable's virtual draw method that takes care of filling
     * and stroking the shape when the fill and/or outline patterns are present.
     */
    virtual void draw( Cairo::RefPtr<Cairo::Context> cairo );

    /**
     * Children should provide their own implementations of this pure virtual
     * method. A child should create the necessary path that will be filled
     * by the draw() method.
     */
    virtual void draw_shape( Cairo::RefPtr<Cairo::Context> cairo ) = 0;

    /**
     * The default implementation calls upon draw_shape() to create the path,
     * but a child should provide their own implementation if the draw_shape()
     * is not sufficient for outlining.
     */
    virtual void draw_outline( Cairo::RefPtr<Cairo::Context> cairo );

    /** The pattern that will be used to fill this shape. */
    Cairo::RefPtr<Cairo::Pattern> fill();

    /** Sets the pattern that will be used to fill this shape. */
    void set_fill( Cairo::RefPtr<Cairo::Pattern> fill = Cairo::RefPtr<Cairo::Pattern>() );

    /** The LineStyle that will be used to draw the outline */
    LineStyle& outline();

    /**
     * Sets the line style that will be used to outline this shape.
     * @param set_use_outline defaults to true and turns on outlining
     */
    void set_outline( const LineStyle& outline, bool use_outline = true );

    /**
     * Sets the line style that will be used to outline this shape.
     * @param set_use_outline defaults to true and turns on outlining
     */
    void set_outline( Cairo::RefPtr<Cairo::Pattern> pattern, bool use_outline = true );

    /** True if an outline will be drawn, false otherwise */
    bool use_outline();

    /**
     * Set to true if an outline should be drawn, false if an outline
     * should not be drawn.
     */
    void set_use_outline( bool use_outline );

    PAPYRUS_CLASS_NAME("Shape");

    protected:
    /**
     * The fill and outline patterns for this shape.
     */
    Cairo::RefPtr<Cairo::Pattern> m_fill;
    LineStyle m_outline;
    bool m_use_outline;
    Cairo::Path* m_fill_path;
    Cairo::Path* m_outline_path;

    sigc::connection m_outline_changed_connection;
    Region m_internal_extents;

    /**
     * Reimplemented to ensure that outline width, if set, is taken into account.
     */
    virtual void set_extents( const Region& extents );

    virtual void on_outline_changed();

    virtual void shape_changed( unsigned which=FILL|OUTLINE );


  };

}

#endif
