/***************************************************************************
 *   Copyright (C) 2004 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#ifndef PAPYRUSTEXT_H
#define PAPYRUSTEXT_H

#include <string>

#include <papyrus/shape.h>

/**
 * \page guide_text Text
 *
 * Text is specified by a size and a string (for now, not all cairo's text
 * handling features are exposed, such as setting the font).
 *
 * A text object is drawn such that the shape's (x,y) coordinate defines the
 * lower-left corner of the text area.
 *
 * The following contains an example that will add a text string to a canvas.
 *
 * The blue lines represent the x and y axes, and show through since an alpha value
 * of 0.9 is used.
 *
 * \image html text.png "Text drawn by example code below"
 *
 * \dontinclude shapes.cpp
 * \skip add_text
 * \until }
 *
 */


namespace Papyrus {

  /**
   * A text object is drawn such that the shape's (x,y) coordinate defines the
   * lower-left corner of the text area.
   *
   * \image html text.png "Text"
   *
   * @author Rick L Vinyard Jr
   */
  class Text : public Shape {
    public:
      typedef PapyrusSmartPointer<Text> pointer;

      Text( const std::string& text = std::string(),
            double size = 12.0,
            const std::string& family = std::string(),
            Cairo::FontSlant slant = Cairo::FONT_SLANT_NORMAL,
            Cairo::FontWeight weight = Cairo::FONT_WEIGHT_NORMAL
          );

      static Text::pointer create( const std::string& text = std::string(),
                                   double size = 12.0,
                                   const std::string& family = std::string(),
                                   Cairo::FontSlant slant = Cairo::FONT_SLANT_NORMAL,
                                   Cairo::FontWeight weight = Cairo::FONT_WEIGHT_NORMAL
                                 );

      ~Text();

      const std::string& font_family();

      Cairo::FontSlant font_slant();

      Cairo::FontWeight font_weight();

      void set_font_face( const std::string& family,
                          Cairo::FontSlant slant,
                          Cairo::FontWeight weight
                        );

      const Cairo::FontOptions& font_options();

      void set_font_options( const Cairo::FontOptions& font_options );

      double size();

      void set_size( double );

      std::string text() const;

      void set_text( const std::string );

      virtual void draw_shape( Cairo::RefPtr<Cairo::Context> cairo );

      /**
       * Reimplemented render method to allow the text extents to be calculated
       * before the bounds of a viewbox are calculated.
       */
      virtual void render( Cairo::RefPtr<Cairo::Context> cairo );

      sigc::signal<void> signal_text();

      sigc::signal<void> signal_font();

      PAPYRUS_CLASS_NAME("Text");

      PAPYRUS_CLONE_METHOD( Text );

		protected:
      std::string m_text;
      double m_size;
      std::string m_family;
      Cairo::FontSlant m_slant;
      Cairo::FontWeight m_weight;
      Cairo::FontOptions m_font_options;

      sigc::signal<void> m_signal_text;

      sigc::signal<void> m_signal_font;
  };

}

#endif
