/***************************************************************************
*   Copyright (C) 2004 by Rick L. Vinyard, Jr.                            *
*   rvinyard@cs.nmsu.edu                                                  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU Lesser General Public License as        *
*   published by the Free Software Foundation version 2.1.                *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU Lesser General Public      *
*   License along with this library; if not, write to the                 *
*   Free Software Foundation, Inc.,                                       *
*   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
***************************************************************************/
#include "viewport.h"

#include <iostream>

#include <cstdlib>
#include <math.h>

#include <assert.h>

namespace PapyrusGtk
{

  Viewport::Viewport() :
      m_scroll_to( SCROLL_TO_CENTER ),
      m_scroll_x( 0.0 ),
      m_scroll_y( 0.0 ),
      m_tie_canvas_size_to_viewport(true),
      m_tie_viewport_size_to_canvas(true)
  {
    m_canvas = Papyrus::Canvas::create();
    m_canvas->set_selectable( false );
    m_canvas_size_connection = m_canvas->signal_size().connect( sigc::mem_fun( *this, &Viewport::on_canvas_size_changed ) );

    m_root = Papyrus::Group::create();
    m_root->set_selectable( false );
    m_canvas_redraw_connection = m_root->signal_redraw().connect( sigc::mem_fun( *this, &Viewport::on_need_redraw ) );
    m_root->add( m_canvas );

    this->signal_expose_event().connect( sigc::mem_fun( *this, &Viewport::expose_event_callback ) );

    this->signal_size_allocate().connect( sigc::mem_fun( *this, &Viewport::size_allocate_callback ) );

    this->signal_button_press_event().connect( sigc::mem_fun( *this, &Viewport::button_press_event_callback ) );

    this->signal_button_release_event().connect( sigc::mem_fun( *this, &Viewport::button_release_event_callback ) );

    this->signal_key_press_event().connect( sigc::mem_fun( *this, &Viewport::key_press_event_callback ) );

    this->signal_key_release_event().connect( sigc::mem_fun( *this, &Viewport::key_release_event_callback ) );

    this->signal_motion_notify_event().connect( sigc::mem_fun( *this, &Viewport::motion_notify_event_callback ) );

    this->signal_scroll_event().connect( sigc::mem_fun( *this, &Viewport::scroll_event_callback ) );

  }

  Viewport::~Viewport()
  {}

  void Viewport::clear()
  {
    m_canvas->clear();
    queue_draw();
  }

  bool Viewport::expose_event_callback( GdkEventExpose* e )
  {
    if ( ! m_context ) {
      cairo_t * cairo;
      /**
      * We need to call the paint function passing it the bounds of the window and the cairo drawing surface.
      * To do this, we need to create the cairo surface using gdk_cairo_create which requires the gdk drawable.
      * But, we can't directly access the gdk drawable from a Gtk::Widget. We must obtain the Gdk::Window of the
      * Gtk::Widget. Gdk::Drawable is a parent of Gdk::Window, but the gdk drawable is only available in the
       * parent method. Therefore we must access it this way:
      *
      * Gtk::Widget -> Gdk::Window -> Gdk::Drawable -> gdk drawable
      *
       */

      // get the Gdk::Window for this Gtk::Widget
      Glib::RefPtr<Gdk::Window> window = get_window();

      if ( window ) {
        // from the Gdk::Window, call the parent Gdk::Drawable gobj() method to obtain the gdk drawable
        GdkDrawable * gdkDrawable = window->Gdk::Drawable::gobj();

        if ( gdkDrawable != NULL ) {
          // create the gdk cairo surface
          cairo = gdk_cairo_create( gdkDrawable );

          m_context = Cairo::RefPtr<Cairo::Context>( new Cairo::Context( cairo, false ) );
        }
      }

    }

    m_root->render( m_context );
    m_context.clear();
    return true;
  }

  void Viewport::on_canvas_size_changed()
  {
    int cw, ch;
    if ( m_tie_viewport_size_to_canvas && ! m_canvas_size_connection.blocked() )
    {
      cw = (int)(ceil( m_canvas->size_w() ));
      ch = (int)(ceil( m_canvas->size_h() ));
      m_canvas_size_connection.block();
      this->set_size_request( cw, ch );
      m_canvas_size_connection.unblock();
    }
  }

  void Viewport::on_need_redraw( double x, double y, double w, double h )
  {
    this->queue_draw_area( ( int ) floor( x ) - 1, ( int ) floor( y ) - 1, ( int ) ceil( w ) + 2, ( int ) ceil( h ) + 2 );
  }

  Papyrus::Canvas::pointer Viewport::canvas( )
  {
    return m_canvas;
  }

  void Viewport::set_canvas( Papyrus::Canvas::pointer canvas )
  {
    if ( m_canvas_size_connection )
      m_canvas_size_connection.disconnect();

    m_root->remove( m_canvas );
    m_canvas = canvas;
    m_root->add( m_canvas );

    m_canvas_size_connection = m_canvas->signal_size().connect( sigc::mem_fun( *this, &Viewport::on_canvas_size_changed ) );
  }

  void Viewport::scroll_to( double x, double y )
  {
    double offset_x, offset_y;
    m_scroll_x = x;
    m_scroll_y = y;
    this->scroll_offsets( offset_x, offset_y );
    m_root->set_translate( x + offset_x, y + offset_y );
  }

  void Viewport::scroll_to_center( )
  {
    scroll_to( 0.0, 0.0 );
  }

  void Viewport::window_to_canvas( double winx, double winy, double & canvasx, double & canvasy ) const
  {
    double offset_x, offset_y;
    this->scroll_offsets( offset_x, offset_y );
    canvasx = m_scroll_x - offset_x + winx;
    canvasy = m_scroll_y - offset_y + winy;
  }

  void Viewport::canvas_to_window( double canvasx, double canvasy, double & winx, double & winy ) const
  {
    double offset_x, offset_y;
    this->scroll_offsets( offset_x, offset_y );
    winx = canvasx - m_scroll_x + offset_x;
    winy = canvasy - m_scroll_y + offset_y;
  }

  void Viewport::set_scroll_point( SCROLL_TO scroll_to )
  {
    if ( scroll_to == m_scroll_to )
      return ;

    m_scroll_to = scroll_to;
    queue_draw();
  }

  SCROLL_TO Viewport::scroll_point( )
  {
    return m_scroll_to;
  }

  void Viewport::scroll_offsets( double & offset_x, double & offset_y ) const
  {
    switch ( m_scroll_to ) {
      case SCROLL_TO_CENTER:
        offset_x = m_width / 2.0;
        offset_y = m_height / 2.0;
        break;
      case SCROLL_TO_TOP_LEFT:
        offset_x = 0.0;
        offset_y = 0.0;
        break;
      case SCROLL_TO_TOP_RIGHT:
        offset_x = m_width;
        offset_y = 0.0;
        break;
      case SCROLL_TO_BOTTOM_LEFT:
        offset_x = 0.0;
        offset_y = m_height;
        break;
      case SCROLL_TO_BOTTOM_RIGHT:
        offset_x = m_width;
        offset_y = m_height;
        break;
    }
  }

  void Viewport::size_allocate_callback( Gtk::Allocation & allocation )
  {
    m_width = allocation.get_width();
    m_height = allocation.get_height();
    if ( m_tie_canvas_size_to_viewport && ! m_canvas_size_connection.blocked() )
    {
      m_canvas_size_connection.block();
      m_canvas->set_size( m_width, m_height );
      m_canvas_size_connection.unblock();
    }
    scroll_to( m_scroll_x, m_scroll_y );
  }

  const Viewport::Controllers& Viewport::controllers()
  {
    return m_controllers;
  }

  void Viewport::add_controller( Papyrus::Controller::pointer controller )
  {
    m_controllers.insert( controller );
  }

  void Viewport::remove_controller( Papyrus::Controller::pointer controller )
  {
    m_controllers.erase( controller );
  }

  bool Viewport::button_press_event_callback( GdkEventButton * event )
  {
    Papyrus::Event::ButtonPress press_event;
    Papyrus::Event::ButtonDoublePress double_press_event;
    Papyrus::Event::ButtonTriplePress triple_press_event;
    Controllers::iterator i;
    double x, y;
    bool handled = false;

    if ( m_controllers.size() > 0 ) {

      this->window_to_canvas( event->x, event->y, x, y );

      switch ( event->type ) {
        case GDK_BUTTON_PRESS:
          press_event = Papyrus::Event::ButtonPress( event->time, event->state, event->button, x, y );
          for ( i = m_controllers.begin(); i != m_controllers.end(); i++ )
            if ( ( *i ) ->handle( press_event ) )
              handled = true;
          break;
        case GDK_2BUTTON_PRESS:
          double_press_event = Papyrus::Event::ButtonDoublePress( event->time, event->state, event->button, x, y );
          for ( i = m_controllers.begin(); i != m_controllers.end(); i++ )
            if ( ( *i ) ->handle( double_press_event ) )
              handled = true;
          break;
        case GDK_3BUTTON_PRESS:
          triple_press_event = Papyrus::Event::ButtonTriplePress( event->time, event->state, event->button, x, y );
          for ( i = m_controllers.begin(); i != m_controllers.end(); i++ )
            if ( ( *i ) ->handle( triple_press_event ) )
              handled = true;
          break;
      }

      if ( handled )
        return true;

    }

    return false;
  }

  bool Viewport::button_release_event_callback( GdkEventButton * event )
  {
    Papyrus::Event::ButtonRelease release_event;
    Controllers::iterator i;
    double x, y;
    bool handled = false;

    if ( m_controllers.size() > 0 ) {

      this->window_to_canvas( event->x, event->y, x, y );

      release_event = Papyrus::Event::ButtonRelease( event->time, event->state, event->button, x, y );

      for ( i = m_controllers.begin(); i != m_controllers.end(); i++ )
        if ( ( *i ) ->handle( release_event ) )
          handled = true;

      if ( handled )
        return true;

    }

    return false;
  }

  bool Viewport::key_press_event_callback( GdkEventKey * event )
  {
    Papyrus::Event::KeyPress press_event;
    Controllers::iterator i;
    bool handled = false;

    if ( m_controllers.size() > 0 ) {

      press_event = Papyrus::Event::KeyPress( event->time, event->state, event->keyval, event->hardware_keycode, event->group );

      for ( i = m_controllers.begin(); i != m_controllers.end(); i++ )
        if ( ( *i ) ->handle( press_event ) )
          handled = true;

      if ( handled )
        return true;

    }

    return false;
  }

  bool Viewport::key_release_event_callback( GdkEventKey * event )
  {
    Papyrus::Event::KeyRelease release_event;
    Controllers::iterator i;
    bool handled = false;

    if ( m_controllers.size() > 0 ) {

      release_event = Papyrus::Event::KeyRelease( event->time, event->state, event->keyval, event->hardware_keycode, event->group );

      for ( i = m_controllers.begin(); i != m_controllers.end(); i++ )
        if ( ( *i ) ->handle( release_event ) )
          handled = true;

      if ( handled )
        return true;

    }

    return false;
  }

  bool Viewport::motion_notify_event_callback( GdkEventMotion * event )
  {
    Papyrus::Event::Motion motion_event;
    Controllers::iterator i;
    double x, y;
    bool handled = false;

    if ( m_controllers.size() > 0 ) {

      this->window_to_canvas( event->x, event->y, x, y );

      motion_event = Papyrus::Event::Motion( event->time, event->state, x, y );

      for ( i = m_controllers.begin(); i != m_controllers.end(); i++ )
        if ( ( *i ) ->handle( motion_event ) )
          handled = true;

      if ( handled )
        return true;

    }

    return false;
  }

  bool Viewport::scroll_event_callback( GdkEventScroll * event )
  {
    Papyrus::Event::Scroll scroll_event;
    Papyrus::Event::ScrollDirection direction;
    Controllers::iterator i;
    double x, y;
    bool handled = false;

    if ( m_controllers.size() > 0 ) {

      this->window_to_canvas( event->x, event->y, x, y );

      switch ( event->direction ) {
        case GDK_SCROLL_UP:
          direction = Papyrus::Event::SCROLL_UP;
          break;
        case GDK_SCROLL_DOWN:
          direction = Papyrus::Event::SCROLL_DOWN;
          break;
        case GDK_SCROLL_LEFT:
          direction = Papyrus::Event::SCROLL_LEFT;
          break;
        case GDK_SCROLL_RIGHT:
          direction = Papyrus::Event::SCROLL_RIGHT;
          break;
      }

      scroll_event = Papyrus::Event::Scroll( event->time, event->state, direction, x, y );

      for ( i = m_controllers.begin(); i != m_controllers.end(); i++ )
        if ( ( *i ) ->handle( scroll_event ) )
          handled = true;

      if ( handled )
        return true;

    }

    return false;
  }

bool Viewport::canvas_size_tied_to_viewport( )
{
  return m_tie_canvas_size_to_viewport;
}

void Viewport::tie_canvas_size_to_viewport( bool tie )
{
  if ( tie == m_tie_canvas_size_to_viewport )
    return;

  m_tie_canvas_size_to_viewport = tie;

  if ( m_tie_canvas_size_to_viewport )
  {
    int w, h;
    this->get_size_request( w, h );
    m_canvas->set_size( w, h );
  }
}

bool Viewport::viewport_size_tied_to_canvas( )
{
  return m_tie_viewport_size_to_canvas;
}

void Viewport::tie_viewport_size_to_canvas( bool tie )
{
  if ( tie == m_tie_viewport_size_to_canvas )
    return;

  m_tie_viewport_size_to_canvas = tie;

  if ( m_tie_canvas_size_to_viewport )
  {
    if ( m_canvas_size_connection )
      m_canvas_size_connection.disconnect();

    m_canvas_size_connection = m_canvas->signal_size().connect( sigc::mem_fun( *this, &Viewport::on_canvas_size_changed ) );

    this->on_canvas_size_changed();
  }
  else
  {
    if ( m_canvas_size_connection )
      m_canvas_size_connection.disconnect();
  }
}

}
